#encoding=utf8
import json
from datetime import timedelta
from uuid import uuid4
from flask_redis import FlaskRedis
from flask import request,session
from redis import Redis
from werkzeug.datastructures import CallbackDict
from flask.sessions import SessionInterface, SessionMixin

class RedisSession(CallbackDict, SessionMixin):
    def __init__(self, initial=None, sid=None, new=False):
        def on_update(self):
            self.modified = True
            self.new = False
        CallbackDict.__init__(self, initial, on_update)
        self.sid = sid
        self.new = new
        self.modified = False
        self.seend_cookie = True
        # 某些临时session的过期时间会比较短或长
        self.temp_expiration_time = 0

class RedisSessionInterface(SessionInterface):
    session_class = RedisSession
    def __get_sid(self,sid):
        if self.prefix:
            return self.prefix +':'+sid
        else:
            return sid

    def __init__(self, app, redis, prefix='session'):
        # if redis is None:
        #     redis = FlaskRedis(strict=False)
        assert redis is not None,'redis 必须要指定'
        self.redis = redis
        # assert self.redis.provider_class == Redis,"请使用提示格式创建rds:rds = FlaskRedis(app,strict=False)"
        self.prefix = prefix
        if app:
            self.init_app(app)

    def init_app(self,app):
        app.session_interface = self
        # self.redis.init_app(app)
        app.session_cookie_name='sessionid'

    def generate_sid(self):
        return str(uuid4())

    def get_redis_expiration_time(self, app, session):
        if session.permanent:
            return app.permanent_session_lifetime
        if session.temp_expiration_time>0:
            return timedelta(seconds=session.temp_expiration_time)
        return timedelta(days=2)

    def get_session(self,sid):
        if not sid:
            sid = self.generate_sid()
            return self.session_class(sid=sid, new=True)
        val = self.redis.get(self.__get_sid(sid))
        if val is not None:
            data = json._default_decoder.decode(val.decode())
            return self.session_class(data, sid=sid)
        return self.session_class(sid=sid, new=True)

    def open_session(self, app, request):
        sid = request.cookies.get(app.session_cookie_name)
        return self.get_session(sid)

    def save_session(self, app, session, response):
        domain = self.get_cookie_domain(app)
        if not session:
            # 如果是新创建的session，还没有写入redis时，不用做delete redis的动作，以提升性能
            if not session.new:
                self.redis.delete(self.__get_sid(session.sid))
            if session.modified:
                response.delete_cookie(app.session_cookie_name,
                                       domain=domain)
            return
        redis_exp = self.get_redis_expiration_time(app, session)
        cookie_exp = self.get_expiration_time(app, session)
        val = json._default_encoder.encode((dict(session)))

        try:
            self.redis.setex(self.__get_sid(session.sid),
                             int(redis_exp.total_seconds()),
                             val
                             )
        except Exception as e:
            # assert self.redis.provider_class == Redis,"请使用提示格式创建rds:rds = FlaskRedis(app,strict=False)"
            self.redis.setex(self.__get_sid(session.sid),val,
                                 int(redis_exp.total_seconds())
                                 )
        # 有某些认证只需要记录session，不需要送给前端
        # 默认每次都送保持，在使用时，保证cookie不过期
        if session.seend_cookie:
            response.set_cookie(app.session_cookie_name,
                                session.sid,
                                expires=cookie_exp,
                                httponly=True,
                                domain=domain)
            # 改为在login中指定，以后每次sendcookie时，不能覆盖掉前端设定的locale
            # response.set_cookie('locale', get_lang(session.get('locale')),
            #                     expires=cookie_exp,
            #                     # 需要让前端修改
            #                     httponly=False,
            #                     domain=domain
            #                     )
            # 确保下次正常时需要
            session.seend_cookie = False
            session.temp_expiration_time = 0



# encoding=utf8
'''
用于微服务的框架，没有认证的url统一返回401
session 存 redis 中，没有User_Class
'''
from functools import wraps
from flask import  session,  g,  make_response
from mwauth.base_auth import BaseAuth,User

class SessionAuth(BaseAuth):
    """docstring for ClassName"""
    app = None
    def __init__(self, app=None):
        super(SessionAuth, self).__init__()
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app
        app.auth = self

    def valid_login(self, func):
        @wraps(func)
        def _request_login(*args, **kw):
            user_id = session.get('uid')
            if user_id:
                g.user_name = session.get('uname')
                g.user_id = session.get('uid')
                g.current_user = User(uid=session.get('uid'),uname=session.get('uname'),
                                      systemuser=session.get('systemuser',False),manageuser=session.get('manageuser',False),
                                      manageuserid=session.get('manageuserid'))
                return func(*args, **kw)
            # 没有认证则返回401
            response = make_response()
            response.status_code = 401
            return response
        return _request_login





class BaseUser(object):
    def get_password(self):
        pass
    def check(self,password):
        pass
    def make(self,password):
        pass
    def to_json(self):
        pass
    @staticmethod
    def FindUser(username):
        pass


##################################
#提供基于api gateway kong的JWT认证和cookie认证
#认证失败时返回401
#################################
from mwauth.base_auth import BaseAuth,User
from flask import request,make_response,g,current_app,session
from flask_babel import _
from functools import wraps
import inspect
from .redis_session import RedisSessionInterface
import hashlib
class KongAuth(BaseAuth):
    # '''
    # 1，在request前提取 user info
    # '''
    def __set_auth(self):
        '''
        保存用户，jwt等auth info
        :return:
        '''
        if not isinstance(current_app.session_interface,RedisSessionInterface):
            raise Exception('请在app/__init__.py的create_app_swagger方法中增加下列代码：\n app.session_interface = RedisSessionInterface(app, rds)')
        # redis 中session的过期时间，为0时，为第一次设定的时间
        session_expiration_time = current_app.config.get('SESSION_EXPIRATION_TIME', 0)
        if current_app.config.get('DEVELOPMENT', False):
            g.user_name = current_app.config.get('LOGIN_USER_NAME')  # 'user_dev'
            g.user_id = current_app.config.get('LOGIN_USER_ID')
            g.current_user = User(uid=current_app.config.get('LOGIN_USER_ID'),
                                  uname=current_app.config.get('LOGIN_USER_NAME'),
                                  systemuser=current_app.config.get('LOGIN_USER_SYSTEMUSER', False),
                                  manageuser=current_app.config.get('LOGIN_USER_MANAGEUSER', False),
                                  manageuserid=current_app.config.get('LOGIN_USER_MANAGEUSER_ID'),
                                  companyid=current_app.config.get('LOGIN_USER_COMPANYID',''),
                                  phone = current_app.config.get('LOGIN_USER_PHONE'),
                                  type=session.get('type', 'appuser')
                                  )
            g.jwt = session.sid
            # 只有session 認證才需要設定這個redis的有效期
            session.temp_expiration_time = session_expiration_time
            session.update({'uid':g.current_user.uid,'uname':g.current_user.uname,
                            'systemuser':g.current_user.systemuser,
                            'manageuser':g.current_user.manageuser,
                            'manageuserid':g.current_user.manageuserid})
            return
        # 通过了jwt 或 key 认证的 api 一定会回传username和userid，否则视为kong的认证不成功
        g.user_name = request.headers.get('X-Consumer-Username')
        g.user_id = request.headers.get('X-Consumer-Custom-Id')
        # key auth 或jwt auth
        g.jwt = request.args.get('jwt') or request.args.get('apikey')or request.args.get('sessionid') or request.args.get('token')
        if g.jwt is None:
            # jwt header authorization: bearer jwt...
            jwt = request.headers.get('authorization', None)
            if jwt:
                # 去掉前面的bearer
                g.jwt = jwt[7:]
            elif session:
                g.jwt = session.sid
                # session 认证可能不经过kong，但也需保证auth 成功，确保代码兼容
                g.user_name = session.get('uname')
                g.user_id = session.get('uid')
            else:
                # header，query，和cookie中都没有 jwt or sessionid or token or apikey，需要重新认证
                g.user_name = None
                g.user_id = None
                return
            # 只有session 認證才需要設定這個redis的有效期
            session.temp_expiration_time = session_expiration_time
        elif session and session.sid!=g.jwt:
            # 如果有传jwt ，就使用jwt的，cookie中的session可能是上一次的，需刷新为jwt的
            session.sid = g.jwt
            mysession = current_app.session_interface.get_session(g.jwt)
            # 只有session 認證才需要設定這個redis的有效期
            session.temp_expiration_time = session_expiration_time
            session.update(dict(mysession))
        if not session:
            # 当非session认证时，还需保存session，保证app中的代码的兼容性
            session.sid = g.jwt
            # 临时session 不需要送cookie，避免浪费流量
            session.seend_cookie = False
            mysession = current_app.session_interface.get_session(g.jwt)
            # 如果只通过了Kong 认证，且没有在redis中存该session key ，如：key auth，则写入kong 认证的信息给session
            if request.headers.get('X-Consumer-Username') \
                and request.headers.get('X-Consumer-Custom-Id') \
                and not mysession:
                # 有 X-Consumer-Username 可能是JWT和key auth, 没有mysession 只有key auth 认证时，才需要更新session
                if request.args.get('apikey')==g.jwt :
                    session.update(dict({'uid':g.user_id ,'uname':g.user_name,
                                'systemuser':False,
                                'manageuser':False,
                                'manageuserid':'',
                                'type':'appuser'}))
                else: # logout 后redis session已经被清除, 但jwt 还在, 通过jwt传递参数的api,需要重新认证
                    g.user_name = None
                    g.user_id = None
                    return
            else:
                # 只有session 認證才需要設定這個redis的有效期
                # session.temp_expiration_time = session_expiration_time
                session.update(dict(mysession))
        g.current_user = User(uid=session.get('uid'),
                              uname=session.get('uname'),
                              systemuser=session.get('systemuser', False),
                              manageuser=session.get('manageuser', False),
                              manageuserid=session.get('manageuserid'),
                              companyid=session.get('companyid',''),
                              type=session.get('type','appuser'),
                              phone=session.get('phone','')
                )
        g.user_name = session.get('uname')
        g.user_id = session.get('uid')


    def __init__(self, app=None):
        super(KongAuth, self).__init__()
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app
        app.auth = self
        # self.app.register_blueprint(auth)

    def valid_login(self, func):
        @wraps(func)
        def _request_login(*args, **kw):
            # 给g 赋 用户等信息
            self.__set_auth()
            if g.user_name or g.user_id :
                return func(*args, **kw)
            # 没有认证则返回401
            response = make_response()
            response.status_code = 401
            return response
        return _request_login


    def valid_sign(self,func):
        @wraps(func)
        def _request_sign(*args, **kw):
            # key认证的token为key，浏览器端因访问不到jwt，可以传username
            # json.dumps(json_body)+token(或用户名)+noncestr+timestamp 转md5
            sign = request.headers.get('X-MW-Sign') or request.args.get("sign")
            # 资料库不区分大小写，会导致前后sign 不一致
            sign_source_str = f'{request.data.decode()}{g.current_user.uname.lower()}' \
                              f'{request.headers.get("X-MW-Noncestr") or request.args.get("noncestr", "")}' \
                              f'{request.headers.get("X-MW-Timestamp") or request.args.get("timestamp", "")}'
            sign_md5 = hashlib.md5(sign_source_str.encode()).hexdigest()
            current_app.logger.info(f'sign_source_user:{sign_source_str}, sign_md5:{sign_md5}, sign_web:{sign}')
            if sign_md5 == sign:
                return func(*args, **kw)
            else:
                sign_source_str = f'{request.data.decode()}{g.jwt}' \
                                  f'{request.headers.get("X-MW-Noncestr") or request.args.get("noncestr", "")}' \
                                  f'{request.headers.get("X-MW-Timestamp") or request.args.get("timestamp", "")}'
                sign_md5_jwt = hashlib.md5(sign_source_str.encode()).hexdigest()
                current_app.logger.info(f'sign_source_jwt:{sign_source_str}, sign_md5_jwt:{sign_md5_jwt},sign_web:{sign}')
                if sign_md5_jwt == sign:
                    return func(*args, **kw)
            # 没有认证则返回400
            response = make_response({ "error": f'sign error,web sign:{sign},'
                                                f'srv sign_user {sign_md5},'
                                                f'srv sign_jwt{sign_md5_jwt},'
                                                f'noncestr:{request.headers.get("X-MW-Noncestr") or request.args.get("noncestr", "")},'
                                                f'timestamp:{request.headers.get("X-MW-Timestamp") or request.args.get("timestamp", "")}'},
                                     400, {"content-type": "application/json;charset=utf8"})
            response.status_code = 400
            return response
        return _request_sign

    def valid_phone_code(self,phone_code_arg_name:str):
        '''
        检查验证码，需要传入验证码的参数名称
        :param phone_code_arg_name: 函数中电话验证码的参数名称
        :return:
        '''
        def wrapper(func):
            @wraps(func)
            def valid_phone_code(*args, **kwargs):
                # 根据phone_code的参数名称获取phone_code验证码
                phone_code = inspect.signature(func).bind_partial(*args, **kwargs).arguments.get(phone_code_arg_name)
                assert phone_code is not None, f'验证码的参数名称{phone_code_arg_name}在函数中不存在'
                phone = g.current_user.phone
                rds = current_app.session_interface.redis
                valid_code, hits = rds.hmget(f'session:phone.valid.code:{phone}', ['valid_code', 'hits'])
                if valid_code:
                    self.max_hits = rds.hincrby(f'session:phone.valid.code:{phone}', 'max_hits', 1)
                if valid_code and valid_code.decode() == phone_code:
                    rds.delete(f'session:phone.valid.code:{phone}')
                    return func(*args, **kwargs)
                rds.hincrby(f'session:phone.valid.code:{phone}', 'hits', -1)
                if hits and int(hits.decode()) <= 0:
                    rds.delete(f'session:phone.valid.code:{phone}')
                return  make_response({ "error": _('The phone valid code is error,please retry enter.')},  400, {"content-type": "application/json;charset=utf8"})
            return valid_phone_code
        return wrapper





from setuptools import setup, find_packages
from codecs import open
import os

def read(f):
    return open(os.path.join(os.path.dirname(__file__), f),encoding='utf8').read().strip()

setup(
    # $ pip install mwauth
    # And where it will live on PyPI: https://pypi.org/project/mwauth/
    name='mwauth',  # Required
    version='0.4.51',  # Required
    description='maxwin auth',  # Required
    # long_description='\n\n'.join((read('README.rst'), read('CHANGES.txt'))),
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='https://bitbucket.org/maxwin-inc/auth/src',  # Optional
    author='cxhjet',  # Optional
    author_email='cxhjet@qq.com',  # Optional
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    keywords='maxwin commonlib auth',  # Optional
    packages=find_packages(exclude=['test','test.*']),  # Required
    install_requires=['flask>=0.11.1'],  # Optional
    include_package_data=True,
    project_urls={  # Optional
        'Bug Reports':'https://bitbucket.org/maxwin-inc/auth/issues?status=new&status=open',
        'Funding': 'https://donate.pypi.org',
        'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://bitbucket.org/maxwin-inc/auth/src',
    },
)



from enum import Enum

from flask import g

from mwauth import KongAuth, SessionAuth


class AuthType(Enum):
    '''
    该类型支持程序启动时，选择auth 方式
    '''
    # kong_auth.KongAuth,同时支持jwt和session认证方式
    jwt = 'jwt'
    # session_auth3.SessionAuth，只支持非kong的cookie认证
    session = 'session'


def auth(auth_type=AuthType.jwt.value, app=None):
    '''
    支持在程序启动的时候，选择认证类型
    :param auth_type: 认证类型，比如：jwt，session...
    :param app:
    :return:
    '''
    if auth_type== AuthType.jwt.value:
        return KongAuth(app)
    elif auth_type== AuthType.session.value:
        return SessionAuth(app)
    else:
        assert False,'auth_type(%s) is not supoort!'%auth_type


def current_user():
    return g.user


# from mwauth import AuthType
# from mwauth.utils import AuthType
from .kong_auth import KongAuth
from .session_auth3 import SessionAuth
from .baseuser import BaseUser
from .utils import AuthType,auth,current_user


